MODULE cantest;

SFR CSR; IR; eXP01C; BTR; GMS; UMLM; LMLM;
    P3; DP3; S0BG; S0CON;
    S0RIC; S0TIC; PSW;
    
CONST AdCSR = $EF00; (* Startadresse des CAN-Bereichs *)
TYPE canobj = RECORD
       (* StandardId im Klartext, MOBnr * 256 + Dlen *)
       id, nr_len: INTEGER;
       D0, D1, D2, D3: INTEGER; (* 8 Datenbytes *)
     END;

VAR Rmob, Tmob: canobj;
    canstatus: INTEGER; (* Bit0: Fehler, sonst Empfgang auf MOB_Bit *)
    i: INTEGER;
    
    (* Standard-CAN mit 125 kBaud bei 20 MHz, s. AP29000 *)
    PROCEDURE* CAN_init;
    VAR i, mob: INTEGER;
    BEGIN i := $0041; CSR := i; (* Starte Init *)
      i := $4944; BTR := i;  (* Setze Timing *)
      i := $E0FF; GMS := i; (* Standardframe ohne Vorfilterung *)
      i := $FFFF; UMLM := i; (* MOB15 ohne Vorfilterung *)
      i := $F8FF; LMLM := i;
      (* Alle MOBs invalid setzen: *)
      mob := AdCSR + 16; i := $FF7F;
      WHILE mob <= (AdCSR + $F0) DO
        VERBATIM
	  mov	[R14], R15
	END_VERBATIM
	mob := mob + 16;
      END;
      (* Error- und allgemein Interrupt enablen *)
      i := $000A; CSR := i; (* Beende Init *);
      (* Interrupts anablen Level 4, Grplev. 2 *)
      eXP01C := %01010010;
      canstatus := 0;
    END CAN_init;
    
    (* Interruptbehandlung fr Empfangs- und Fehler-Interrupts.
       Status- und Sende-Interrupts sollten nicht freigegeben werden. *)
    PROCEDURE* ISR[64]; (* CAN-Interrupt *)
    VAR status, i, j, mob: INTEGER;
    BEGIN status := IR; i := CSR; 
      j := i; j := j ANDB $00FF; CSR := j; (* Reset Status *)
      IF (i ANDB $C000) # 0 THEN (* CAN-ERROR *)
	canstatus := canstatus ORB 1;
	(* ev. BOff und EWRN extra merken? *)
      ELSE (* Empfangsint. *)
        i := status; i := i MOD 256; 
        IF i = 2 THEN canstatus := canstatus ORB $8000;
	  mob := AdCSR + $F0; (* MOB15-Message-Control *)
	  (* ev. Spezialbehandlung fr MOB15-Empfang *)
        ELSE (* Empfang durch MOB1 ... MOB14 *)
          i := i - 2; j := 1;
          VERBATIM
	    shl	R13, R14	; j := 2^i
          END_VERBATIM
          canstatus := canstatus ORB j;
          mob := AdCSR + i * 16;
        END;
        i := $FFFD; (* Reset INTPND im Message-Control-Register *)
        VERBATIM
	  mov	[R12], R14
        END_VERBATIM
      END;      
    END ISR;
           
    (* Tmob senden *)
    PROCEDURE* CAN_send(remote: BOOLEAN);
    VAR mob, i, D0, D1, D2, D3, D4: INTEGER;
    BEGIN i := CSR; CSR := i ANDB $F7FF; (* Reset TXOK *)
      mob := Tmob.nr_len; mob := mob DIV 256; mob := mob * 16 + AdCSR;
      D0 := $5A55; (* setze CPUUPD und NEWDAT, andere Flags auf 0  *)
      i := Tmob.id; i := i MOD 8; i := i * (32 * 256);  D1 := i;
      i := Tmob.id; i := i DIV 8; D1 := D1 + i; (* ID in D1 *)
      D2 := 0;
      VERBATIM
	mov	[R14], R12		; D0 nach MCRn 
	mov	[R14+2], R11		; ID nach UAR_Mn
	mov	[R14+4], R10		; 0 nach LAR_Mn
      END_VERBATIM
      D0 := Tmob.D0; D0 := D0 ANDB $00FF; D0 := D0 * 256;
      i := Tmob.nr_len; i := i ANDB $00FF; D0 := D0 + (i * 16) + 8;
      D1 := Tmob.D0; D1 := D1 DIV 256; i := Tmob.D1; i := i * 256; D1 := D1 + i;
      D2 := Tmob.D1; D2 := D2 DIV 256; i := Tmob.D2; i := i * 256; D2 := D2 + i;
      D3 := Tmob.D2; D3 := D3 DIV 256; i := Tmob.D3; i := i * 256; D3 := D3 + i;
      D4 := Tmob.D3; D4 := D4 DIV 256; 
      (* CPUUPD = 0, MSGVAL = 1; TXRQ nach remote, andere Flags unveraendert *)
      IF remote THEN i := $F79F
      ELSE i := $C79F; END;
      VERBATIM
	mov	[R14+6], R12		; D0 sichern
	mov	[R14+8], R11		; D1 sichern
	mov	[R14+10], R10		; D2 sichern
	mov	[R14+12], R9		; D3 sichern
	mov	[R14+14], R8		; D4 sichern
	mov	[R14], R13		; i nach MCRn
      END_VERBATIM
    END CAN_send;
    
    (* Empfangsobjekt anlegen *)
    PROCEDURE* CAN_receive(mob, id: INTEGER; remote: BOOLEAN);
    VAR i, conf: INTEGER;
    BEGIN mob := mob * 16 + AdCSR; 
      i := id; i := i MOD 8; i := i * (32 * 256); id := (id DIV 8) + i;
      conf := 0; (* DLC = 0, DIR = 0, XTD = 0 *)
      IF remote THEN i := $6599
      ELSE i := $5599; END;
      VERBATIM        
	mov	[R15+#2], R14		; UARn := id
	mov	[R15+#4], R11		; LARn := conf (= 0)
	mov	[R15+#6], R11		; MCFGn := conf
	mov	[R15], R12		; MCRn := i
      END_VERBATIM
    END CAN_receive;
            
    (* empfangenes Mob in Rmob kopieren *)
    PROCEDURE* ReadMob(mob: INTEGER);
    VAR i, id, D0, D1, D2, D3, D4: INTEGER;
    BEGIN mob := mob * 256; Rmob.nr_len := mob;
      mob := mob DIV 16 + AdCSR;
      VERBATIM
	mov	R13, [R15+#2]		; id := UARn
	mov	R12, [R15+#6]		; D0 := d0/Conf
	mov	R11, [R15+#8]		; D1 := d2/d1
	mov	R10, [R15+#10]		; D2 := d4/d3
	mov	R9, [R15+#12]		; D3 := d6/d5
	mov	R8, [R15+#14]		; D5 := d7
      END_VERBATIM
      i := id; i := i DIV (32 * 256); id := id ANDB $00FF; 
      id := id * 8 + i; Rmob.id := id;
      i := D0; i := i ANDB $00F0; i := i DIV 16 + Rmob.nr_len;
      Rmob.nr_len := i;
      D0 := D0 DIV 256; i := D1; i := i * 256; Rmob.D0 := i + D0;
      D1 := D1 DIV 256; i := D2; i := i * 256; Rmob.D1 := i + D1;
      D2 := D2 DIV 256; i := D3; i := i * 256; Rmob.D2 := i + D2;
      D3 := D3 DIV 256; i := D4; i := i * 256; Rmob.D3 := i + D3;
    END ReadMob;
    
    (* Status eines Mobs auslsen, CSR und MCRn landen in nr *)
    PROCEDURE* MobStat(VAR nr: INTEGER); (* nr = 1 .. 14 *)
    VAR mob, i, j, val, t: INTEGER;
    BEGIN mob := nr; mob := mob * 16 + AdCSR;
      VERBATIM
	mov	R14, [R14]		; mob := MCRn
      END_VERBATIM 
      j := 1; val := 0;
      FOR i := 1 TO 8 DO
	t := mob;
	IF t MOD 2 = 0 THEN
	  val := val + j; END;
	j := j * 2; mob := mob DIV 4;
	(*VERBATIM
            jb     R14.0, MobStat1	; mob ungerade?
            add    R11, R12		; nein: val := val + j
          MobSTAT1:
            shl    R12, #1		; j := j * 2
            shr    R14, #2		; mob := mob DIV 4
        END_VERBATIM*)
      END;
      mob := CSR; mob := mob ANDB $FF00;
      nr := mob + val;      
    END MobStat;
    
    (* Mob als ungueltig erklaeren *)
    PROCEDURE* MobInval(nr: INTEGER);
    VAR i: INTEGER;
    BEGIN nr := nr * 16 + AdCSR; i := $FF7F;
      VERBATIM
	mov	[R15], R14
      END_VERBATIM
    END MobInval;        
    
    PROCEDURE* IniASC0;
    VAR i: INTEGER;
    BEGIN i := DP3; DP3 := i ORB 1024; (* P3.10 = TxD, output *)
      i := P3; P3 := i ORB 1024; (* P3.10 = TxD = 1 *)
      S0BG := 15; (* 38400 Baud bei 20 MHz Takt *)
      S0RIC := 0; S0TIC := 0; (* keine Interrupts hierdurch *)
      S0CON := $8011; (* 8N1 *)
    END IniASC0;

    PROCEDURE* test1;    
    VAR i, j: INTEGER;
    BEGIN Tmob.id := $0401; Tmob.D0 := (97 * 256) + 72; (* Ha *)
      Tmob.D1 := (108 * 256) + 108; (* ll *)
      Tmob.D2 := (33 * 256) + 111; (* o! *)
      Tmob.nr_len := 256 + 6; (* MOB1, 6 Bytes *)
      CAN_send(FALSE);
    END test1;
    
    PROCEDURE* test2;
    SFR Pled = P1L; DPLED = eDP1L; 
    VAR i: INTEGER;
    BEGIN CAN_receive(2, $0402, FALSE); 
      DPLED := 255; (* 8 outputs *) Pled := 254;
      i := 0;
      WHILE i < 256 DO
	Sleep;
	IF (canstatus ANDB  %100) # 0 THEN
	  canstatus := canstatus ANDB ($FFFF - %100);
	  ReadMob(2);
	  i := Rmob.D0 MOD 256; Pled := i XORB 255;
	END;
      END;    
    END test2;

BEGIN IniASC0; CAN_init; 
  WriteLn("CAN initialisiert.");
  i := PSW; PSW := i + 2048; (* IEN = 1 *)
  test1; test2;
  REPEAT
    Sleep;
    WriteLn("Status: ", canstatus:h);
    canstatus := 0;
  UNTIL canstatus = $FFFF;
END cantest.
